/**************************************************************************
**  (c) Copyright 2002, Andromeda Technology & Automation
** This is free software; you can redistribute it and/or modify it under the
** terms of the GNU General Public License, see the file COPYING.
***************************************************************************
** MODULE INFORMATION *
***********************
**      FILE NAME      : gcm_input.cpp
**      SYSTEM NAME    : Gnucomo - Gnu Computer Monitoring
**      VERSION NUMBER : $Revision: 1.10 $
**
**  DESCRIPTION      :  Application to store client messages into the database
**                      The client message contains a log file from one of the
**                      system logs. Gcm_input parses the log file and enters
**                      the raw data into the 'log' table of the gnucomo database.
**
**                      The system log may arrive in one of several forms:
**                      1. Log file (archived or not) from a client machine.
**                      2. Log file from a client system, arriving in a clear Email.
**                      3. Log file from a client machine, arriving in an
**                         encrypted Email.
**
**                      additional information we need that may not be availble in
**                      the content of the log is:
**                      - FQDN of the client.
**                      - Time of arrival of the log
**                      - Service that created the log.
**
**                      If the log arrives in an Email, these items probably are
**                      available in the mail header. Otherwise, they have to be
**                      provided as command line arguments.
**
**                      Command line arguments:
**                      -c <name>        Configuration name (default = gnucomo).
**                      -d <date>        Date and time of log arrival.
**                      -h <hostname>    FQDN of the client.
**                      -i               Incremental - partial list of parameters.
**                      -s <service>     Service that created the log.
**                      -T               Test mode. Do not alter the database.
**                      -v               Verbose (debug) output.
**                      -V               Print version and exit.
**
**  EXPORTED OBJECTS : 
**  LOCAL    OBJECTS : 
**  MODULES  USED    :
***************************************************************************
**  ADMINISTRATIVE INFORMATION *
********************************
**      ORIGINAL AUTHOR : Arjen Baart - arjen@andromeda.nl
**      CREATION DATE   : Aug 29, 2002
**      LAST UPDATE     : Aug 11, 2003
**      MODIFICATIONS   : 
**************************************************************************/

/*****************************
   $Log: gcm_input.cpp,v $
   Revision 1.10  2003/09/03 06:58:31  arjen
   Changed version string to 0.0.8

   Revision 1.9  2003/09/01 06:59:26  arjen
   A date without the time for the '-d <date> option will
   assume midnight on that date.

   Revision 1.8  2003/08/16 15:30:19  arjen
   Fixed a gcc 2 vs. gcc 3 problem

   Revision 1.7  2003/08/14 10:28:37  arjen
   Use parameters from a new section 'logging' with three configuration parameters:
      method       - Output method to use for logging.
      destination  - Name of the log output destination.
      level        - Log level: Verbose output if greater than 0.

   Revision 1.6  2003/08/11 16:56:16  arjen
   Different kinds of log files are parsed by a collection of objects
   of different classes, derived from the base class line_cooker
   Depending on the message content or the message_type element in
   XML, one of these objects is selected.

   Logrunner is integrated with gcm_input. Although its functionality
   is still limited, a connection between logrunner and gcm_input
   is beginning to form.

   Revision 1.5  2003/08/05 08:11:06  arjen
   Added two configuration parameters:
      logfile   - Log to this file instead of stderr.
      verbosity - Verbose output if greater than 0.
   Added '-i' option for incremental parameter updates

   Revision 1.4  2003/03/29 08:42:00  arjen
   Exit without reading any input if the database connection fails.

   Revision 1.3  2002/11/09 08:04:27  arjen
   Added a reference to the GPL

   Revision 1.2  2002/11/04 10:13:36  arjen
   Use proper namespace for iostream classes

   Revision 1.1  2002/10/05 10:25:49  arjen
   Creation of gcm_input and a first approach to a web interface

*****************************/

static const char *RCSID = "$Id: gcm_input.cpp,v 1.10 2003/09/03 06:58:31 arjen Exp $";

#include <fstream>

#include <getopt.h>

#include "message.h"
#include "syslog_cooker.h"
#include "irix_syslog_cooker.h"
#include "access_cooker.h"
#include "error_cooker.h"

bool verbose = false;
bool testmode = false;
bool incremental = false;
std::ostream *log = &std::cerr;

static char *Version = "gcm_input version 0.0.8 - Sep 04, 2003";


/*=========================================================================
**  NAME           : main
**  SYNOPSIS       : int main(int argc, char *argv[])
**  PARAMETERS     : 
**  RETURN VALUE   : 
**
**  DESCRIPTION    : Parse command line arguments and establish a connection
**                   to the database.
**                   When we have a database connection, parse the log file
**                   from stdin.
**
**  VARS USED      :
**  VARS CHANGED   :
**  FUNCTIONS USED :
**  SEE ALSO       :
**  LAST MODIFIED  : Aug 11, 2003
**=========================================================================
*/

int main(int argc, char *argv[])
{
   const char *usage = "Usage: gcm_input [-c configname] [-h hostname] [-i] [-d date]"
                       " [-s service] [-T] [-v] [-V]\n";

   gnucomo_config    cfg;
   char             *config_name = "gnucomo";
   std::ofstream    logfile;


   /*   Parse command line arguments */

   UTC    arrival = Now();
   String  hostname(""), service("");
   int     option;


   while ((option = getopt(argc, argv, "c:h:d:s:iTvV")) != -1)
   {
      switch (option)
      {
      case 'c':
         config_name = optarg;
         break;

      case 'h':
         hostname = optarg;
         break;

      case 'd':
         arrival = String(optarg);
         if (!date(arrival).proper())
         {
            std::cerr << "gcm_input: Invalid date string: " << optarg
                      << "(" << arrival << ").\n";
            exit(1);
         }
         else if (!hour(arrival).proper())
         {
            arrival = UTC(date(arrival), hour(0,0,0));
         }
         break;

      case 's':
         service = optarg;
         break;

      case 'i':
         incremental = true;
         break;

      case 'T':
         testmode = true;
         break;

      case 'v':
         verbose = true;
         break;

      case 'V':
         std::cout << Version << "\n";
         exit(0);

      case '?':
      case ':':
         std::cerr << usage;
         exit(1);
      }
   }
   /*  Get the configuration file */

   if (!cfg.read(config_name))
   {
      std::cerr << "Can not read Gnucomo configuration file for " << config_name << ".\n";
      exit(1);
   }

   String log_method      = cfg.find_parameter("logging", "method");
   String log_destination = cfg.find_parameter("logging", "destination");
   int    level           = cfg.find_parameter("gcm_input", "level");

   if (log_method == "file" && log_destination != "")
   {
      std::cerr << "Logging to " << log_destination << ".\n";
#if __GNUC__ == 2
      logfile.open(log_destination, _IO_APPEND); // for gcc 2
#else
      logfile.open(log_destination, std::ios_base::app); // for gcc 3
#endif
      if (!logfile)
      {
         std::cerr << "Can't open logfile " << log_destination << " for writing.\n";
      }
      else
      {
         log = &logfile;
         verbose = verbose || level > 0;
      }
   }

   if (verbose)
   {
      *log << "Hostname = " << hostname;
      *log << " Arrival = " << arrival;
      *log << " Service = " << service << "\n";
      *log << "Config OK.\n";
   }

   /*  Try to connect to the database */

   gnucomo_database db(&cfg);

   if (db.is_connected())
   {

      client_message      msg(&std::cin, db);
      syslog_cooker       slc;
      irix_syslog_cooker  islc;
      access_cooker       alc;
      error_cooker        elc;

      msg.add_cooker(&slc);
      msg.add_cooker(&islc);
      msg.add_cooker(&alc);
      msg.add_cooker(&elc);

      if (msg.classify(hostname, arrival, service) > 0.9)
      {
         msg.enter();
      }
      return 0;
   }
   else
   {
      *log << "gcm_input: Can not connect to database.\n";
      *log << "Gcm_input finished at " << Now() << ".\n";
      return 1;
   }
}

