/*
 * logrunner.c
 * (c) Peter Roozemaal, feb 2003
 *
 * $Id: logrunner.cpp,v 1.1 2003/08/11 16:56:16 arjen Exp $
 *
 * 1) compile,
 * 2) create 'logrunner.conf' containing a list of filenames and types of the
 *    logfiles to scan; one file per line:
 *       <gnucomo_messagetype> TAB <logfilename>
 *    example:
 *      syslog   /var/log/messages
 *      syslog   /var/log/secure
 * 3) run executable
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <stdio.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>

#include <AXE/String.h>

extern String XML_Entities(String s);

static const char* const usage =
   "Usage: logrunner [<options>] [<hostname>]\n\n"
   "<hostname> is the gnucomo data collection server\n"
   "Options are:\n"
   " -1: one-shot; stop after a single pass over one of the configured logfiles\n"
   " -p <number>: set IP port number to connect to\n"
   " -c <file>: specify alternative configuration file, default is \'logrunner.conf\'\n"
   "If no hostname is specified, logrunner sends output to stdout\n";

struct fileinfo {
   struct fileinfo* next;
   char*   name;      /* strdupped name */
   char*   type;      /* strdupped logtype */
   int     fd;
   ino_t   inode;
   off_t   position;
};

static const char* confname = "logrunner.conf";
static const char* statusname = "logrunner.status";
static const char* newstatusname = "logrunner.status.new";

static const char *hostname = NULL;
static int port = 2996;         /* random magic number */
static int out_stream = 1;

static int oneshot = 0;

static struct fileinfo* filelist = NULL;

static int something_has_changed = 0;
   
static volatile sig_atomic_t sig_seen = 0;

static void sighandler(int sig)
{
   sig_seen = sig;
}

static void set_signal_handler(void)
{
   /* These aren't all; but it's a nice set to start with */
   signal(SIGHUP, sighandler);
   signal(SIGINT, sighandler);
   signal(SIGQUIT, sighandler);
   signal(SIGABRT, sighandler);
   signal(SIGPIPE, sighandler);
   signal(SIGTERM, sighandler);
}

void open_output()
{
   struct hostent* hostptr;
   struct sockaddr_in addr;
   unsigned int namelen = sizeof(addr);

   if ( hostname )
   {
      hostptr = gethostbyname(hostname);
      if ( !hostptr )
      {
         fprintf(stderr, "logrunner: FATAL: cannot resolve %s\n", hostname);
         exit(2);
      }
      out_stream = socket(PF_INET, SOCK_STREAM, 0);
      if ( out_stream < 0 )
      {
         fprintf(stderr, "logrunner: FATAL: Socket creation failed\n");
         exit(2);
      }
      addr.sin_family = AF_INET;
      addr.sin_addr.s_addr = *((long*)(hostptr->h_addr));
      addr.sin_port = htons(port);
      if ( connect(out_stream, (struct sockaddr*) &addr, namelen) < 0 )
      {
         fprintf(stderr, "logrunner: FATAL: connect to %s failed\n", hostname);
         exit(2);
      }
   }
   else
   {
      out_stream = 1;
   }
}

void xsend(const char* str)
{
   write(out_stream, str, strlen(str));
}

void xml_header(struct fileinfo *f)
{
   char buffer[256];
   *buffer = 0;
   gethostname(buffer, sizeof(buffer));
        xsend("<?xml version='1.0'?>\n");
   xsend("<gcmt:message xmlns:gcmt=\"http://gnucomo.org/transport/\">\n");
   xsend("<gcmt:header>\n<gcmt:hostname>");
   xsend(buffer);
   xsend("</gcmt:hostname>\n<gcmt:messagetype>");
   xsend(f->type);
   xsend("</gcmt:messagetype>\n</gcmt:header>\n");
   xsend("<gcmt:data><gcmt:log>");
}

void xml_footer(void)
{
   xsend("</gcmt:log></gcmt:data>\n</gcmt:message>\n");
}

void output(const char* str)
{
   write(2, str, strlen(str));
}

void output_error(const char* str)
{
   char error[80];
   sprintf(error, " errno=%d\n", errno);
   output(str);
   output(error);
}

void write_status_file(struct fileinfo* fl)
{
   FILE* dumpfile;
   int localerror = 0;
   dumpfile = fopen(newstatusname, "w");
   if ( !dumpfile )
   {
      char error[80];
      sprintf(error, ", errno=%d\n", errno);
      output("!!! dumpstatus: open failed: ");
      output(newstatusname);
      output(error);
      something_has_changed = 0;
      return;
   }
   while ( fl )
   {
      if ( fprintf(dumpfile, "%s %li %lu\n",
              fl->name, (long) fl->inode,
              (unsigned long) fl->position) < 0 )
      {
         output_error("!!! dumpstatus: write failed");
         localerror = 1;
         break;
      }
      fl = fl->next;
   }
   if ( fclose(dumpfile) && localerror == 0 )
   {
      output_error("!!! dumpstatus: close failed");
      localerror = 1;
   }
   if ( localerror == 0 )
   {
      if ( rename(newstatusname, statusname) )
      {
         output_error("!!! dumpstatus: rename failed");
      }
   }
   something_has_changed = 0;
}

void read_status()
{
   FILE* statusfile;
   char buffer[4096];
   char* xp;
   long ino;
   unsigned long pos;
   struct fileinfo* fp;

   statusfile = fopen(statusname, "r");
   if ( statusfile == NULL )
   {
      fprintf(stderr, "logrunner: can\'t open status file \'%s\': ",
         statusname);
      perror("");
      return;
   }
   while ( fgets(buffer, sizeof(buffer), statusfile ) )
   {
      xp = strchr(buffer, ' ');
      if ( xp )
      {
         *xp = 0;
         sscanf(xp+1, "%ld %lu", &ino, &pos);
         fp = filelist;
         while ( fp )
         {
            if ( strcmp(buffer, fp->name) == 0 )
            {
               fp->inode = ino;
               fp->position = pos;
               break;
            }
            fp = fp->next;
         }
      }
   }
   fclose(statusfile);
}

struct fileinfo* new_info()
{
   struct fileinfo* rv;
   rv = (struct fileinfo *)malloc(sizeof(struct fileinfo));
   if ( rv )
   {
      rv->next = NULL;
      rv->name = NULL;
      rv->type = NULL;
      rv->fd = -1;
      rv->position = 0;
      rv->inode = 0;
   }
   return rv;
}

void read_config()
{
   /* The config file now is a simple list of type, filenames;
    * it should be more user-friendly in the future!
    */
   FILE* conffile;
   char buffer[4096];

   conffile = fopen(confname, "r");
   if ( !conffile )
   {
      fprintf(stderr,
         "logrunner: can\'t open configuration file \'%s\': ",
         confname);
      perror("");
      exit(2);
   }
   while ( fgets(buffer, sizeof(buffer), conffile) )
   {
      int len = strlen(buffer);
      char *fname;
      if ( buffer[len-1] == '\n' )
      {
         buffer[--len] = 0;
      }
      fname = strchr(buffer, '\t');
      if ( fname )
      {
         struct fileinfo* fi;
         fi = new_info();
         *(fname++) = 0;
         if ( fi )
         {
            fi->name = strdup(fname);
            fi->type = strdup(buffer);
            fi->next = filelist;
            filelist = fi;
         }
         else
         {
            fprintf(stderr, "logrunner: out of memory\n");
            exit(2);
         }
      }
      else if ( len > 0 )
      {
         fprintf(stderr, "logrunner: WARNING: bad config line: %s\n", buffer);
      }
   }
   fclose(conffile);
}

void copy_data(struct fileinfo *f)
{
   char buffer[4096];
   int ndata;

   /* read data and dump to output */
   ndata = read(f->fd, buffer, sizeof(buffer));
   if ( ndata > 0 )
   {
      xml_header(f);
      while ( ndata > 0 )
      {
         //  Make a separate <gcmt:raw> element from each line
 
         char *line, *nextline;

         line = buffer;
         nextline = buffer;

         while (nextline < buffer + ndata)
         {
            while (*nextline != '\n' && nextline < buffer + ndata)
            {
               nextline++;
            }
            if (*nextline == '\n')
            {
               // Another line found - make the split.
               *nextline++ = '\0';
               write(out_stream, "<gcmt:raw>", 10);

               String logline(line);
               logline = XML_Entities(logline);
               write(out_stream, logline, ~logline);
               write(out_stream, "</gcmt:raw>\n", 12);
               line = nextline;
            }
         }
         if (line != nextline)
         {
            //  There is still an incomplete line in the buffer.
            memmove(buffer, line, nextline - line);
         }
         f->position += ndata - (nextline - line);
         ndata -= line - buffer;
         ndata += read(f->fd, buffer + (nextline - line), sizeof(buffer) - (nextline - line));
      }
      xml_footer();
      something_has_changed = 1;
   }   
   if ( ndata < 0 )
   {
      char error[80];
      sprintf(error, ", errno=%d\n", errno);
      output("!!! logfile: read failed: ");
      output(f->name);
      output(error);
   }
}

void do_file(struct fileinfo *f)
{
   struct stat statinfo;
   char buffer[80];

   /* inode check */
   if ( stat(f->name, &statinfo) )
   {
      sprintf(buffer, ", errno=%d\n", errno);
      output("!!! logfile: stat failed: ");
      output(f->name);
      output(buffer);
   }
   else
   {
      if ( statinfo.st_ino != f->inode )
      {
         if ( f->fd >= 0 )
         {
            copy_data(f);
            close(f->fd);
            f->fd = -1;
         }
         output("@@@ logfile: logrotate detected: ");
         output(f->name);
         output("\n");
         f->inode = statinfo.st_ino;
         f->position = 0;
      }
   }
   if ( f->fd < 0 )
   {
      /* attempt to open the file */
      f->fd = open(f->name, O_RDONLY);
      if ( f->fd < 0 )
      {
         sprintf(buffer, ", errno=%d\n", errno);
         output("!!! logfile: open failed: ");
         output(f->name);
         output(buffer);
         return;
      }
      output("*** logfile: opened: ");
      output(f->name);
      sprintf(buffer,
         "\n*** logfile: resumed read from position %ld\n",
         (long) lseek(f->fd, f->position, SEEK_SET) );
      output(buffer);
   }

   copy_data(f);
}

void process_options(int argc, char* argv[])
{
   const char* const options = "1c:p:";
   int opt;

   opt = getopt(argc, argv, options);
   while ( opt != -1 )
   {
      switch ( opt )
      {
      case '1':
         oneshot = 1;
         break;
      case 'c':
         confname = strdup(optarg);
         break;
      case 'p':
         port = atoi(optarg);
         break;
      default:
         fputs(usage, stderr);
         exit(2);
      }
      opt = getopt(argc, argv, options);
   }
   switch ( argc-optind )
   {
   case 0:
      break;
   case 1:
      hostname = argv[optind];
      break;
   default:
      fputs(usage, stderr);
      exit(2);
   }
}

int main(int argc, char* argv[])
{
   process_options(argc, argv);
   open_output();
   
   read_config();
   read_status();

   set_signal_handler();

   while ( sig_seen == 0 )
   {
      struct fileinfo* fip;
      fip = filelist;
      while ( fip )
      {
         do_file(fip);
         if (something_has_changed && oneshot)
         {
            break;
         }
         fip = fip->next;
      }
      if ( something_has_changed )
      {
         write_status_file(filelist);
      }
      if ( oneshot )
      {
         return 0;
      }
      sleep(1);
   }

   fprintf(stderr, "logrunner: stopped by signal %d\n", sig_seen);
   /* shouldn't we close files and release memory here? */
   return 0;
}
