
/**************************************************************************
**  (c) Copyright 2002, Andromeda Technology & Automation
** This is free software; you can redistribute it and/or modify it under the
** terms of the GNU General Public License, see the file COPYING.
***************************************************************************
** MODULE INFORMATION *
***********************
**      FILE NAME      : message.cpp
**      SYSTEM NAME    : Gnucomo - Gnu Computer Monitoring
**      VERSION NUMBER : $Revision: 1.14 $
**
**  DESCRIPTION      :  Implementation of the message handling classes
**
**  EXPORTED OBJECTS : 
**  LOCAL    OBJECTS : 
**  MODULES  USED    :
***************************************************************************
**  ADMINISTRATIVE INFORMATION *
********************************
**      ORIGINAL AUTHOR : Arjen Baart - arjen@andromeda.nl
**      CREATION DATE   : Sep 16, 2002
**      LAST UPDATE     : Jul 24, 2003
**      MODIFICATIONS   : 
**************************************************************************/

/*****************************
   $Log: message.cpp,v $
   Revision 1.14  2003/09/01 06:57:14  arjen
   Reject log entries that are found to be invalid.

   Revision 1.13  2003/08/16 15:28:45  arjen
   Fixed a namespace problem

   Revision 1.12  2003/08/11 16:56:16  arjen
   Different kinds of log files are parsed by a collection of objects
   of different classes, derived from the base class line_cooker
   Depending on the message content or the message_type element in
   XML, one of these objects is selected.

   Logrunner is integrated with gcm_input. Although its functionality
   is still limited, a connection between logrunner and gcm_input
   is beginning to form.

   Revision 1.11  2003/08/05 08:15:00  arjen
   Debug output to the log stream instead of cerr.
   Fixed namespace problems in XPath searches of the DOM.
   Moved string utility functions to a separate file.

   Revision 1.10  2003/04/29 09:16:44  arjen
   Read XML input,
   Only cooked log entries for now.

   Revision 1.9  2003/03/29 09:04:10  arjen
   Extract the hostname out of the 'From:' or 'Message-Id:' line
   of an email header.

   Revision 1.8  2003/03/16 09:42:40  arjen
   Read IRIX system logs.

   Revision 1.7  2003/02/21 08:08:05  arjen
   Gcm_input also detects packages that are removed from the system.
   Determining the version number of a package in a RPM
   list is improved. Only the last one or two parts of the string that
   begin with a '-' and a number are considered the version.

   Revision 1.6  2003/02/05 09:37:51  arjen
   Create notifications when a new package is discovered
   in a 'rpm -qa' list or when the version of a package is changed.

   Revision 1.4  2002/12/06 22:26:28  arjen
   Set the value of log.processed to FALSE when inserting a
   new log entry into the database
   When a syslog entry arrives from last year, gcm_input subtracts one from the
   year of arrival to create the year of the log entry.
   Read output from "rpm -qa" and enter packages in the parameter table.

   Revision 1.3  2002/11/09 08:04:27  arjen
   Added a reference to the GPL

   Revision 1.2  2002/11/04 10:13:36  arjen
   Use proper namespace for iostream classes

   Revision 1.1  2002/10/05 10:25:49  arjen
   Creation of gcm_input and a first approach to a web interface

*****************************/

static const char *RCSID = "$Id: message.cpp,v 1.14 2003/09/01 06:57:14 arjen Exp $";

#include <algorithm>
#include <libxml/xpath.h>
#include <libxml/debugXML.h>
#include "message.h"

//#define DEBUG

extern bool verbose;   /*  Defined in the main application */
extern bool testmode;
extern bool incremental;
extern std::ostream *log;

/*   Utility functions   */

extern String SQL_Escape(String s);

/*=========================================================================
**  NAME           : operator >>
**  SYNOPSIS       : bool operator >> (message_buffer &, String &)
**  PARAMETERS     : 
**  RETURN VALUE   : True if input was available.
**
**  DESCRIPTION    : Input operator. Read the next line from the message.
**
**  VARS USED      :
**  VARS CHANGED   :
**  FUNCTIONS USED :
**  SEE ALSO       :
**  LAST MODIFIED  : Nov 04, 2002
**=========================================================================
*/

bool operator >> (message_buffer &b, String &s)
{
   bool   input_ok = false;

   if (b.next_line == b.buffer.end())
   {
      String   l;

      if (*(b.input) >> l)
      {
         b.buffer.push_back(l);

         //   next_line keeps pointing to the end.

         s = l;
         input_ok = true;
      }
   }
   else
   {
      s = *(b.next_line);
      b.next_line++;
      input_ok = true;
   }
   return input_ok;
}

client_message::client_message(std::istream *in, gnucomo_database db)
{
   input.from(in);
   database = db;

   hostname = "";
   mail_header   = false;
   gpg_encrypted = false;
   classification = UNKNOWN;
   xmlDom         = NULL;
   certainty      = 0.0;
}

static const String mail_date_re("[[:alpha:]]{3}, [ 123]?[0-9] [[:alpha:]]{3} [0-9]{4} [0-9]{2}:[0-9]{2}:[0-9]{2} [+-][0-9]{4}");
static const String unix_date_re("[[:alpha:]]{3} [[:alpha:]]{3} [ 123][0-9] [0-9]{2}:[0-9]{2}:[0-9]{2} [0-9]{4}");

static const regex re_PGP("-----BEGIN PGP MESSAGE-----");
static const regex re_dump("^ *DUMP: Date of this level");
static const regex re_rpm("[[:alnum:]+-]+-[0-9][[:alnum:].-]");

static const regex re_uxmail_from("^From [^ \t]+[ ]+" + unix_date_re);
static const regex re_mail_From("^From:[[:blank:]]+");
static const regex re_mail_Date("^Date:[[:blank:]]+" + mail_date_re);
static const regex re_mail_MsId("^Message-Id:[[:blank:]]+");
static const regex re_email_address("[[:alnum:]_.-]+@[[:alnum:]_.-]+");
static const regex re_email_user("[[:alnum:]_.-]+@");
static const regex re_xml_header("xml .*\?>$");

/*=========================================================================
**  NAME           : readXMLinput
**  SYNOPSIS       : int readXMLinput(String first_line)
**  PARAMETERS     : 
**  RETURN VALUE   : Parse the XML input and extract the header information
**
**  DESCRIPTION    : 
**
**  VARS USED      :
**  VARS CHANGED   :
**  FUNCTIONS USED :
**  SEE ALSO       :
**  LAST MODIFIED  : Jul 24, 2003
**=========================================================================
*/

int client_message::readXMLinput(String first_line)
{
   xmlParserCtxtPtr ctxt;
   String           line;
   xmlNodePtr       root, item;
   xmlNsPtr         namespaces[1];

   xmlXPathObjectPtr res;
   xmlXPathContextPtr pathcontext;


   ctxt = xmlCreatePushParserCtxt(NULL, NULL, first_line, ~first_line, NULL);
   while (input >> line)
   {
      xmlParseChunk(ctxt, line, ~line, 0);
   }
   xmlParseChunk(ctxt, "", 0, 1);
   xmlDom = ctxt->myDoc;
   xmlFreeParserCtxt(ctxt);

   root = xmlDocGetRootElement(xmlDom);
   namespaces[0] = root->ns;

   //TODO Ought to check  root->name and  root->ns->href 

   pathcontext = xmlXPathNewContext(xmlDom);
   pathcontext->node = xmlDocGetRootElement(xmlDom);
   pathcontext->namespaces = namespaces;
   pathcontext->nsNr       = 1;

#ifdef DEBUG
   xmlDebugDumpNodeList(stdout, pathcontext->node, 0);
#endif

   res = xmlXPathEval((const xmlChar *)"gcmt:header/gcmt:messagetype/text()", pathcontext);
   if (res->nodesetval != NULL)
   {
#ifdef DEBUG
      xmlDebugDumpNodeList(stdout, *res->nodesetval->nodeTab, 0);
#endif
      item = *res->nodesetval->nodeTab;

      //  Select a line cooker based on the message type.

#ifdef DEBUG
      std::cout << "Looking for a line cooker for " << item->content << "\n";
#endif
      std::list<line_cooker *>::iterator lci = kitchen.begin();
      pan = 0;
      while (pan == 0 && lci != kitchen.end())
      {
         pan = *lci;
         if (pan->message_type() != (const char *)(item->content))
         {
            pan = 0;
         }
         lci++;
      }
      if (pan == 0)
      {
         *log << "Can not find a line cooker for message type " << item->content << "\n";
      }
   }
   else
   {
      *log << "Message type not found in XML header.\n";
   }

   res = xmlXPathEval((const xmlChar *)"gcmt:header/gcmt:hostname/text()", pathcontext);
   if (res->nodesetval != NULL)
   {
#ifdef DEBUG
      xmlDebugDumpNodeList(stdout, *res->nodesetval->nodeTab, 0);
#endif
      item = *res->nodesetval->nodeTab;
      hostname = (const char *)item->content;
   }
   else
   {
      *log << "Hostname not found in XML header.\n";
   }

   res = xmlXPathEval((const xmlChar *)"gcmt:header/gcmt:service/text()", pathcontext);
   if (res->nodesetval != NULL)
   {
      item = *res->nodesetval->nodeTab;
      service = (const char *)item->content;
   }
   res = xmlXPathEval((const xmlChar *)"gcmt:header/gcmt:time/text()", pathcontext);
   if (res->nodesetval != NULL)
   {
      item = *res->nodesetval->nodeTab;
      arrival = String((char *)item->content);

   }
   //xmlDebugDumpNodeList(stdout, *res->nodesetval->nodeTab, 0);

}

/*=========================================================================
**  NAME           : classify
**  SYNOPSIS       : double classify(String host, date arriv_d, hour arriv_t, String serv)
**  PARAMETERS     : 
**  RETURN VALUE   : The certainty with which the message is classified.
**
**  DESCRIPTION    : 
**
**  VARS USED      :
**  VARS CHANGED   :
**  FUNCTIONS USED :
**  SEE ALSO       :
**  LAST MODIFIED  : Aug 11, 2003
**=========================================================================
*/

double client_message::classify(String host, UTC arriv, String serv)
{
   String    line;

   hostname    = host;
   arrival     = arriv;
   service     = serv;

   /*  First, check if the message has a mail header. */

   if (input >> line && line == re_uxmail_from)
   {
      String    from_address;

      mail_header = true;

      /*  Scan ahead for the hostname and date of arrival.  */

      while (input >> line && line != "")
      {
         if (line == re_mail_From)
         {
            from_address = line(re_email_address);
            from_address(re_email_user) = "";            //  Remove the user part;
            if (from_address != "" && ~hostname < ~from_address)
            {
               *log << "Detected hostname " << from_address << "\n";
               hostname = from_address;
            }
         }
         if (line == re_mail_MsId)
         {
            from_address = line(re_email_address);
            from_address(re_email_user) = "";            //  Remove the user part;
            if (from_address != "" && ~hostname < ~from_address)
            {
               *log << "Detected hostname " << from_address << "\n";
               hostname = from_address;
            }
         }
         if (line == re_mail_Date)
         {
            arrival = UTC(line(regex(mail_date_re)));
         }
      }
   }
   else
   {
      //  Push the first line back, we need to read it again.
      --input;

   }

   pan = 0;

   /*
    *  Now that we have the mail header out of the way, try to figure
    *  out what the content of the message is.
    */


   while (input >> line && certainty < 0.9)
   {
      if (verbose)
      {
         *log << "  testing: " << line << "\n";
      }

      if (line == re_xml_header)
      {
         certainty = 1.0;
         classification = XML;
         if (verbose)
         {
            *log << "XML input detected.\n";
         }
         readXMLinput(line);
      }
      else if (line == re_PGP)
      {
         certainty = 1.0;
         gpg_encrypted = true;
         *log << "The message is PGP/GnuPG encrypted.\n";
      }
      else if (line == re_dump)
      {
          certainty = 1.0;
          if (verbose)
          {
             *log << "DUMP output detected.\n";
          }
      }
      else if (line == re_rpm)
      {
          certainty = 1.0;
          classification = RPMLIST;
          service = "";
          if (verbose)
          {
             *log << "RPM package list detected.\n";
          }
      }
      else
      {
         //  Scan the list of line cookers if there is anything familiar.

         std::list<line_cooker *>::iterator lci = kitchen.begin();
         pan = 0;
         while (pan == 0 && lci != kitchen.end())
         {
            pan = *lci;
            if (!pan->check_pattern(line))
            {
               pan = 0;
            }
            lci++;
         }
         if (pan != 0)
         {
            certainty = 1.0;
            classification = COOKER_OBJECT;
            if (verbose)
            {
               *log << "Detected message type " << pan->message_type() << "\n";
            }
         }
      }
   }
   input.rewind();

   if (hostname == "")
   {
      *log <<  "Can not determine the hostname where the message came from.\n";
      certainty = 0.0;
   }
   else if (!arrival.proper())
   {
      *log << "Arrival time is not known.\n";
      certainty = 0.0;
   }
   else
   {
      certainty = 1.0;
   }

   return certainty;
}

/*=========================================================================
**  NAME           : enterXML
**  SYNOPSIS       : int enterXML()
**  PARAMETERS     : 
**  RETURN VALUE   : None
**
**  DESCRIPTION    : Analyze the DOM tree from the XML input.
**                   The DOM tree was previously parsed by readXMLinput().
**
**  VARS USED      :
**  VARS CHANGED   :
**  FUNCTIONS USED :
**  SEE ALSO       :
**  LAST MODIFIED  : Jul 24, 2003
**=========================================================================
*/

void client_message::enterXML()
{
   xmlXPathObjectPtr res;
   xmlXPathContextPtr pathcontext;
   xmlNsPtr           namespaces[1];

   /*  Try to find the host in the database */

   String objectid;

   objectid = database.find_host(hostname);
   if (objectid == "")
   {
      *log << "Please define the host " << hostname << " in the database.\n";
      return;
   }
   if (verbose)
   {
      *log << "Object id for " << hostname << " is " << objectid << "\n";
   }

   pathcontext = xmlXPathNewContext(xmlDom);
   pathcontext->node = xmlDocGetRootElement(xmlDom);
   namespaces[0] = pathcontext->node->ns;
   pathcontext->namespaces = namespaces;
   pathcontext->nsNr       = 1;

   res = xmlXPathEval((const xmlChar *)"gcmt:data/node()", pathcontext);

   if (res->nodesetval != NULL)
   {
      //  Find the first child element of the <data> element.

      xmlNodePtr  node = *res->nodesetval->nodeTab;
      while (node->type != XML_ELEMENT_NODE)
      {
         node = node->next;
      }
      if (strcmp((char *)node->name, "log") == 0)
      {
         //  Each child contains a log entry, raw or cooked.

         node = node->children;
         while (node != NULL)
         {
            if (node->type == XML_ELEMENT_NODE)
            {
               xmlNodePtr  item;
               String      log_hostname;
               UTC         log_date;
               String      raw("");;
               String      log_service;

               if (strcmp((char *)node->name, "raw") == 0)
               {
                  item = node->children;
                  if (pan == 0)
                  {
                     *log << "Can not cook this type of <raw> log element.\n";
                  }
                  else
                  {
                     raw = String((const char *)item->content);
                     if (pan->cook_this(raw, arrival))
                     {
                        log_hostname = pan->hostname();
                        if (log_hostname == "")
                        {
                           log_hostname = hostname;
                        }
                        log_service = pan->service();
                        log_date    = pan->timestamp();
                     }
                     else
                     {
                        *log << "Log line " << raw << " does not match.\n";
                        raw = "";
                     }
                  }
               }
               else if (strcmp((char *)node->name, "cooked") == 0)
               {
                  //  Find the parts of the log entry

                  if (verbose)
                  {
                     *log << "Analyzing cooked element.\n";
                  }
                  pathcontext->node = node;

                  res = xmlXPathEval((const xmlChar *)"hostname/text()", pathcontext);
                  if (res->nodesetval != NULL)
                  {
                     item = *res->nodesetval->nodeTab;
                     log_hostname = (const char *)item->content;
                     if (log_hostname != hostname(0, ~log_hostname))
                     {
                        *log << "Hostname " << log_hostname << " does not match.\n";
                        log_hostname = "";
                     }
                  }
                  else
                  {
                     log_hostname = hostname;
                  }

                  res = xmlXPathEval((const xmlChar *)"service/text()", pathcontext);
                  if (res->nodesetval != NULL)
                  {
                     item = *res->nodesetval->nodeTab;
                     log_service = (const char *)item->content;
                  }
                  else
                  {
                     log_service = service;
                  }

                  res = xmlXPathEval((const xmlChar *)"timestamp/text()", pathcontext);
                  if (res->nodesetval != NULL)
                  {
                     item = *res->nodesetval->nodeTab;
                     log_date = String((const char *)item->content);
                  }
                  else
                  {
                     *log << "<timestamp> missing from cooked log element.\n";
                  }

                  res = xmlXPathEval((const xmlChar *)"raw/text()", pathcontext);
                  if (res->nodesetval != NULL)
                  {
                     item = *res->nodesetval->nodeTab;
                     raw = String((const char *)item->content);
                  }
                  else
                  {
                     *log << "<raw> missing from cooked log element.\n";
                  }

               }

               //   Insert a new log record into the database.
               if (raw != "" && log_hostname != "" && log_date.proper())
               {
                  String insertion("insert into log (objectid, servicecode,"
                        " object_timestamp, timestamp, rawdata, processed) values (");

                  /*   Insert a new record into the log table   */

                  insertion += "'" + objectid + "',";
                  insertion += "'" + log_service + "',";
                  insertion += "'" + log_date.format("%Y-%m-%d %T") + "',";
                  insertion += "'" + arrival.format("%Y-%m-%d %T") + "',";
                  insertion += "'" + SQL_Escape(raw) + "',FALSE";
                  insertion += ")";

                  if (testmode)
                  {
                     *log << insertion << "\n";
                  }
                  else
                  {
                     database.Query(insertion);
                  }

                  if (verbose)
                  {
                     *log << "\n\n";
                  }
               }
 
            }
            node = node->next;
         }
      }
      else
      {
         *log << "Data element " << node->name << " is not supported.\n";
      }
   }
   else
   {
      *log << "Data node not found.\n";
   }
}

/*=========================================================================
**  NAME           : enter
**  SYNOPSIS       : int enter()
**  PARAMETERS     : 
**  RETURN VALUE   : The number of lines successfully parsed from the input
**
**  DESCRIPTION    : 
**
**  VARS USED      :
**  VARS CHANGED   :
**  FUNCTIONS USED :
**  SEE ALSO       :
**  LAST MODIFIED  : Jul 24, 2003
**=========================================================================
*/

int client_message::enter()
{
   if (classification == XML)
   {
      enterXML();
      return 1;
   }

   long   nr_lines = 0;
   String line;
   String qry;

   String change_notification("");
   String create_notification("");
   bool initial_entry = false;

   std::list<String>  packages;


   /*  Double-check the classification of the message */

   if (classification == UNKNOWN || certainty < 0.9 || gpg_encrypted)
   {
      return 0;
   }

   if (mail_header)
   {
      //  Skip the mail header.

      while (input >> line && line != "");
   }

   /*  Try to find the host in the database */

   String objectid;

   objectid = database.find_host(hostname);
   if (objectid == "")
   {
      *log << "Please define the host " << hostname << " in the database.\n";
      return 0;
   }
   if (verbose)
   {
      *log << "Object id for " << hostname << " is " << objectid << "\n";
   }

   if (classification == RPMLIST)
   {

      int n_packages;

      /*   Read all packages, so we will know which ones are  */
      /*   missing at the end.                                */

      qry = "select name from parameter where objectid='";
      qry += objectid + "' and class='package'";
      n_packages = database.Query(qry);
      initial_entry = n_packages == 0;

#ifdef DEBUG
      *log << n_packages << " packages in database.\n";
#endif
      for (int t = 0; t < n_packages; t++)
      {
         packages.push_back(database.Field(t, "name"));
      }
#ifdef DEBUG
      *log << "Package list built: " << packages.size() << ".\n";
#endif
   }

   /*  Scan the input line by line, entring records into the database */

   String rest;   //  Rest of the line to be parsed
   regex  re_any(".*");

   while (input >> line)
   {
      if (verbose)
      {
         *log << line << "\n";
      }


      /*  Check each line if it contains valid information */

      const regex *check;

      switch (classification)
      {
      case RPMLIST:
            check = &re_rpm;
            break;
      case COOKER_OBJECT:
            check = &re_any;
            break;
      }

      if (line == *check)
      {
         date   log_date;
         hour   log_time;
         int    i;

         String insertion("insert into log (objectid, servicecode,"
                           " object_timestamp, timestamp, rawdata, processed) values (");
         String datestring;

         switch (classification)
         {
         case COOKER_OBJECT:
#ifdef DEBUG
            std::cerr << "\ncooker check: " << pan->check_pattern(line) << "\n";
#endif
            if (pan->cook_this(line, arrival))
            {
               if (pan->hostname() == hostname(0,~pan->hostname()))
               {

#ifdef DEBUG
                  std::cerr << " Information from cooker:\n";
                  std::cerr << "     timestamp = " << pan->timestamp() << "\n";
                  std::cerr << "     hostname  = " << pan->hostname() << "\n";
                  std::cerr << "     service   = " << pan->service() << "\n";
#endif
                  /*   Insert a new record into the log table   */

                  insertion += "'" + objectid + "',";
                  insertion += "'" + pan->service() + "',";
                  insertion += "'" + pan->timestamp().format("%Y-%m-%d %T") + "',";
                  insertion += "'" + arrival.format("%Y-%m-%d %T") + "',";
                  insertion += "'" + SQL_Escape(line) + "',FALSE";
                  insertion += ")";

                  if (testmode)
                  {
                     *log << insertion << "\n";
                  }
                  else
                  {
                     database.Query(insertion);
                  }

                  if (verbose)
                  {
                     *log << "\n\n";
                  }

                  nr_lines++;
               }
               else
               {
                  *log << "   Hostname " << pan->hostname() << " does not match.\n";
               }
            }
            else
            {
               *log << "gcm_input WARNING: Not a valid line: " << line << "\n";
            }
            break;


         case RPMLIST:
            //  Scan a list of packages and versions from "rpm -a".
            //  A similar listing can be created on IRIX 6.5 by using the
            //  command "showprods -3 -n|awk '{printf "%s-%s\n",$2,$3}'|grep -v '^[-=]' \
            //            |grep -v Version-Description".
            //
            //  We have to separate the package name and the version.
            //  The separation is marked by a '-', followed by a digit.
            //  However, there may be other sequences of '-'digit in the package name,
            //  do we have to scan ahead until there is at most one such sequence
            //  left in the version string. The '-'digit seqeunce inside the
            //  version usually separates the version and the release number.

            int  version_start, next_version_start;

            i = line.index('-');
            version_start = i;
            next_version_start = i;

            while (i < ~line - 1)
            {
               while (i < ~line - 1 && !(line[i] == '-' && isdigit(line[i + 1])))
               {
                  i++;
               }
               if (i < ~line - 1)
               {
                  version_start = next_version_start;
                  next_version_start = i;
               }
               i++;
            }
            
            if (!isdigit(line[version_start + 1]))
            {
               version_start = next_version_start;
            }
            String package(line(0,version_start));
            String version(line(version_start + 1, ~line));
            String paramid;
            String remark;
            String insert_h;

            if (verbose)
            {
               *log << "Package is " << package;
               *log << ", version is " << version << "\n";
            }

            //  Construct a qry to check the package's existance

            qry = "select paramid from parameter where objectid='";
            qry += objectid + "' and class='package' and name='";
            qry += package + "'";

            if (database.Query(qry) == 1)
            {
               std::list<String>::iterator  lp;

               lp = find(packages.begin(), packages.end(), package);
               if (lp != packages.end())
               {
                  packages.erase(lp);
               }
               else
               {
                  *log <<  "Could NOT find " << package << " in list.\n";
               }

               paramid = database.Field(0, "paramid");
               qry = "select value from property where paramid='";
               qry += paramid + "' and name='version'";
               if (database.Query(qry) == 0)
               {
                  *log << "Database corruption: Package " << package;
                  *log << " does not have a 'version' property.\n";
               }
               else if (database.Field(0, "value") != version)
               {
                  if (verbose)
                  {
                     *log << "  Parameter " << package << " has different version\n";
                  }
                  insertion = "update property set value='";
                  insertion += version + "' where paramid='";
                  insertion += paramid + "' and name='version'";

                  insert_h = "insert into history (paramid, modified, change_nature, changed_property, new_value)";
                  insert_h += " values ('";
                  insert_h += paramid + "', '" + arrival.format("%Y-%m-%d %T") + "', 'MODIFIED', 'version', '";
                  insert_h += version + "')";

                  database.Query(insertion);
                  database.Query(insert_h);

                  if (change_notification == "")
                  {
                     remark = "Gnucomo detected a different version for package parameter(s) ";
                     change_notification = database.new_notification(objectid, "property modified", remark);
                  }

                  if (change_notification != "")
                  {
                     insertion = "insert into parameter_notification (notificationid, paramid) values ('";
                     insertion += change_notification + "', '";
                     insertion += paramid + "')";

                     database.Query(insertion);
                  }
                  else
                  {
                     *log << "gcm_input ERROR: Cannot create 'property modified' notification.\n";
                  }
               }
               else
               {
                  if (verbose)
                  {
                     *log << "   Parameter " << package << " has not changed.\n";
                  }
               }
            }
            else
            {

               if (verbose)
               {
                  *log << "  Parameter " << package << " does not exist.\n";
               }
               //  Create a new package parameter, including version property and history record

               insertion = "insert into parameter (objectid, name, class, description) values ('";
               insertion += objectid + "', '" + package + "', 'package', 'RPM package " + package + "')";
               if (testmode)
               {
                  paramid = "0";
                  *log << insertion << "\n";
               }
               else
               {
                  database.Query(insertion);
                  qry = "select paramid from parameter where objectid='";
                  qry += objectid + "' and class='package' and name='";
                  qry += package + "'";
                  database.Query(qry);
                  paramid = database.Field(0, "paramid");
               }

               insertion = "insert into property (paramid, name, value, type) values ('";
               insertion += paramid + "', 'version', '";
               insertion += version + "', 'STATIC')";
               insert_h = "insert into history (paramid, modified, change_nature, changed_property, new_value)";
               insert_h += " values ('";
               insert_h += paramid + "', '" + arrival.format("%Y-%m-%d %T") + "', 'CREATED', 'version', '";
               insert_h += version + "')";

               if (testmode)
               {
                  *log << insertion << "\n" << insert_h << "\n";
               }
               else
               {
                  database.Query(insertion);
                  database.Query(insert_h);
                  if (!initial_entry)
                  {
                     if (create_notification == "")
                     {
                        remark = "Gnucomo detected new parameter(s) of class package";
                        create_notification = database.new_notification(objectid, "parameter created", remark);
                     }
                     if (create_notification != "")
                     {
                        insertion = "insert into parameter_notification (notificationid, paramid) values ('";
                        insertion += create_notification + "', '";
                        insertion += paramid + "')";

                        database.Query(insertion);
                     }
                     else
                     {
                        *log << "gcm_input ERROR: Cannot create 'parameter created' notification.\n";
                     }
                  }
               }
            }

            if (verbose)
            {
               *log << "\n";
            }

            nr_lines++;
            break;

         }
      }
      else
      {
         *log << "gcm_input WARNING: Not a valid line: " << line << "\n";
      }
   }

   if (classification == RPMLIST && !incremental)
   {
      std::list<String>::iterator  lp;
      String     remove_notification("");

      /*
       *     If there are any packages left in the list, they seem to have
       *     disappeared from the system.
       */

      for (lp = packages.begin(); lp != packages.end(); lp++)
      {
         String paramid;
         String remark;
         String insert;

         //  Construct a qry to check the package's existance

         qry = "select paramid from parameter where objectid='";
         qry += objectid + "' and class='package' and name='";
         qry += *lp + "'";

         if (database.Query(qry) == 1)
         {
            paramid = database.Field(0, "paramid");
            qry ="select change_nature from history where paramid='";
            qry += paramid + "' order by modified desc";
            if (database.Query(qry) <= 0)
            {
               *log << "Database ERROR: no history record for parameter " << *lp << ".\n";
            }
            else if (database.Field(0, "change_nature") != "REMOVED")
            {
               if (verbose)
               {
                 *log << "Removing parameter " << *lp << ".\n";
               }

               insert = "insert into history (paramid, modified, change_nature)";
               insert += " values ('";
               insert += paramid + "', '" + arrival.format("%Y-%m-%d %T") + "', 'REMOVED')";

               database.Query(insert);

               if (remove_notification == "")
               {
                  remark = "Gnucomo detected that package(s) have disappeared ";
                  remove_notification = database.new_notification(objectid, "parameter removed", remark);
               }

               if (remove_notification != "")
               {
                  insert = "insert into parameter_notification (notificationid, paramid) values ('";
                  insert += remove_notification + "', '";
                  insert += paramid + "')";

                  database.Query(insert);
               }
               else
               {
                  *log << "gcm_input ERROR: Cannot create 'parameter removed' notification.\n";
               }
            }
         }
      }
   }

   if (verbose)
   {
      *log << nr_lines << " lines parsed from the log file.\n";
   }
   return nr_lines;
}
